"""Test the arts of Uwuification."""

from __future__ import annotations

import pytest

from imsosorry.uwuification import (
    EMOJIS,
    char_replace,
    emoji,
    nyaify,
    stutter,
    tildify,
    uwuify,
    word_replace,
)


@pytest.mark.parametrize(
    ("in_text", "out_text"),
    [
        ("cats are small", "cats are smol"),
        ("I love dogs", "I luv dogs"),
    ],
)
def test_word_replace(in_text: str, out_text: str) -> None:
    """Test replacing words in text."""
    assert word_replace(in_text) == out_text


@pytest.mark.parametrize(
    ("in_text", "out_text"),
    [
        ("look", "wook"),
        ("rook", "wook"),
        ("wrong", "wrong"),
        ("lr", "ww"),
        ("rl", "ww"),
        ("wrl", "wrw"),
        ("rlw", "wlw"),
        ("wrlw", "wrlw"),
    ],
)
def test_char_replace(in_text: str, out_text: str) -> None:
    """Test replacing characters in text."""
    assert char_replace(in_text) == out_text


@pytest.mark.parametrize(
    ("strength", "in_text", "out_text"),
    [
        (0.0, "cats are small", "cats are small"),
        (1.0, "I love dogs", "I l-love d-dogs"),
    ],
)
def test_stutter(strength: float, in_text: str, out_text: str) -> None:
    """Test adding stutters to text."""
    assert stutter(in_text, strength) == out_text


@pytest.mark.parametrize(
    ("in_text", "out_text"),
    [
        ("naa", "naa"),
        ("nan", "nyan"),
    ],
)
def test_nyaify(in_text: str, out_text: str) -> None:
    """Test nyaifying text."""
    assert nyaify(in_text) == out_text


@pytest.mark.parametrize(
    ("strength", "has_emoji"),
    [
        (0.0, False),
        (1.0, True),
    ],
)
def test_emoji(strength: float, has_emoji: bool) -> None:  # noqa: FBT001 - yes
    """Test adding emojis."""
    output = emoji("I love dogs!", strength)
    assert any(emoji_ in output for emoji_ in EMOJIS) == has_emoji


@pytest.mark.parametrize(
    ("strength", "in_text", "out_text"),
    [
        (0.0, "cats are small", "cats are small"),
        (1.0, "I love dogs", "I~ love~ dogs~"),
    ],
)
def test_tildify(strength: float, in_text: str, out_text: str) -> None:
    """Test adding tildes."""
    assert tildify(in_text, strength) == out_text


@pytest.mark.parametrize(
    "text",
    [
        "-",
        "^",
        ")",
    ],
)
def test_recursion(text: str) -> None:
    """Check that recursion doesn't cause an infinite loop."""
    assert uwuify(text) in EMOJIS



"""The ancient arts of Uwuification."""

from __future__ import annotations

import random
import re
from functools import partial

WORD_REPLACE = {
    "small": "smol",
    "cute": "kawaii~",
    "fluff": "floof",
    "love": "luv",
    "stupid": "baka",
    "idiot": "baka",
    "what": "nani",
    "meow": "nya~",
    "roar": "rawrr~",
}
"""A dict to match certain words for replacement words"""

EMOJIS = [
    "rawr x3",
    "OwO",
    "UwU",
    "o.O",
    "-.-",
    ">w<",
    "(⑅˘꒳˘)",
    "(ꈍᴗꈍ)",
    "(˘ω˘)",
    "(U ᵕ U❁)",
    "σωσ",
    "òωó",
    "(///ˬ///✿)",
    "(U ﹏ U)",  # noqa: RUF001 - literally the point...
    "( ͡o ω ͡o )",
    "ʘwʘ",
    ":3",
    ":3",  # important enough to have twice
    "XD",
    "nyaa~~",
    "mya",
    ">_<",
    "😳",
    "🥺",
    "😳😳😳",
    "rawr",
    "^^",
    "^^;;",
    "(ˆ ﻌ ˆ)♡",  # noqa: RUF001 - literally the point...
    "^•ﻌ•^",
    "/(^•ω•^)",
    "(✿oωo)",
]
"""A list of emojis/emoticons to add."""

REGEX_WORD_REPLACE = re.compile(r"(?<!w)[lr](?!w)")
"""A wegex that to detect certain characters to change to "w"s."""

REGEX_PUNCTUATION = re.compile(r"[.!?\r\n\t]")
"""A regex to detect certain punctuation characters to emotify /(^•ω•^)"""

REGEX_TILDE = re.compile(r"(?![^ ])(?<!\B)")
"""A regex to find places to add tildes (~) to."""

REGEX_STUTTER = re.compile(r"(\s)([a-zA-Z])")
"""A regex to find words to stutter."""
SUBSTITUTE_STUTTER = r"\g<1>\g<2>-\g<2>"
"""A regex to add st-stuttering to strings."""

REGEX_NYA = re.compile(r"n([aeou][^aeiou])")
"""A regex to detect words with an n before a vowel to nyaify."""
SUBSTITUTE_NYA = r"ny\1"
"""A regex to to nyaify words."""


def word_replace(text: str) -> str:
    """Replace words that are keys in the word replacement hash to the values specified."""
    for word, replacement in WORD_REPLACE.items():
        text = text.replace(word, replacement)
    return text


def char_replace(text: str) -> str:
    """Replace certain characters with 'w'."""
    return REGEX_WORD_REPLACE.sub("w", text)


def stutter_replace(match: re.Match[str], strength: float = 0.0) -> str:
    """Replace a single character with a stuttered character."""
    match_string = match.group()
    if random.random() < strength:
        # Stutter the last character
        return f"{match_string}-{match_string[-1]}"
    return match_string


def stutter(text: str, strength: float) -> str:
    """Add stuttering to a string."""
    return REGEX_STUTTER.sub(partial(stutter_replace, strength=strength), text, 0)


def nyaify(text: str) -> str:
    """Nyaify a string by adding a 'y' between an 'n' and a vowel."""
    return REGEX_NYA.sub(SUBSTITUTE_NYA, text, 0)


def emoji_replace(match: re.Match[str], strength: float = 0.0) -> str:
    """Replace a punctuation character with an emoticon."""
    match_string = match.group()
    if random.random() < strength:
        return f" {random.choice(EMOJIS)} "
    return match_string


def emoji(text: str, strength: float) -> str:
    """Replace some punctuation with emoticons."""
    return REGEX_PUNCTUATION.sub(partial(emoji_replace, strength=strength), text, 0)


def tildes(match: re.Match[str], strength: float = 0.0) -> str:
    """Add some tildes to spaces."""
    match_string = match.group()
    if random.random() < strength:
        return "~"
    return match_string


def tildify(text: str, strength: float) -> str:
    """Add some tildes to spaces."""
    return REGEX_TILDE.sub(partial(tildes, strength=strength), text, 0)


def uwuify(
    text: str,
    *,
    stutter_strength: float = 0.2,
    emoji_strength: float = 0.1,
    tilde_strength: float = 0.1,
    max_emojifiable_len: int = 2,
) -> str:
    """Take a string and returns an uwuified version of it."""
    alpha = any(char.isalpha() for char in text)

    if len(text) < max_emojifiable_len and not alpha:
        return random.choice(EMOJIS)

    original_text = text.lower()

    text = text.lower()
    text = word_replace(text)
    text = nyaify(text)
    text = char_replace(text)
    text = stutter(text, stutter_strength)
    text = emoji(text, emoji_strength)
    text = tildify(text, tilde_strength)

    if text == original_text and alpha:
        text = uwuify(
            text,
            stutter_strength=stutter_strength + 0.225,
            emoji_strength=emoji_strength + 0.075,
            tilde_strength=tilde_strength + 0.175,
        )

    return text
